#!/usr/bin/env python
# coding: utf-8

# In[ ]:


import random as rnd
import math
import numpy as np
import pandas as pd
from sklearn.metrics import f1_score
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.metrics import accuracy_score, f1_score
import sklearn.gaussian_process as gp
import tensorflow as tf
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from openpyxl import Workbook, load_workbook
import os
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from tensorflow.keras.optimizers.legacy import Adam
from tensorflow.keras.layers import Dropout
from tensorflow.keras.optimizers import RMSprop
from tensorflow.keras.losses import Huber
import time

# Class for utility functions
class Utils:
    @staticmethod
    def calculate_diversity(orcas):
        # Calculate the average distance between orcas to measure diversity
        total_distance = 0
        num_orcas = len(orcas)
        
        for i in range(num_orcas):
            for j in range(i+1, num_orcas):
                # Euclidean distance calculation
                distance = sum([(a - b)**2 for a, b in zip(orcas[i].position, orcas[j].position)])**0.5
                total_distance += distance
                
        # Calculating the average distance
        average_distance = total_distance / (num_orcas * (num_orcas - 1) / 2)
        return average_distance
    
# Class for Deep Q-Learning
class DQL:
    def __init__(
        self, state_size, action_size=40, buffer_size=2000, target_update_freq=50
    ):
        # Initialize state and action dimensions, memory buffer, and other parameters
        self.state_size = state_size
        self.action_size = action_size
        # Experience replay memory
        self.memory = []
        self.buffer_size = buffer_size
        # Discount rate
        self.gamma = 0.95
        # Exploration rate
        self.epsilon = 1.0
        # Minimum exploration rate
        self.epsilon_min = 0.01
        # Epsilon decay rate
        self.epsilon_decay = 0.995
        # Learning rate
        self.learning_rate = 0.001
        # Main Q-learning model
        self.model = self._build_model()
        # Target model for stabilizing Q-value updates
        self.target_model = self._build_model()
        # Frequency of target model update
        self.target_update_freq = target_update_freq
        # Training step counter
        self.train_step = 0

        # Randomly initialized parameters for exploration
        self.a = rnd.uniform(0, 1)
        self.b = rnd.uniform(0, 1)
        self.d = rnd.uniform(0, 1)
        self.e = rnd.uniform(0, 1)

    def _build_model(self):
        # Build neural model with additional layers and Dropout for regularization
        model = Sequential()
        model.add(Dense(24, input_dim=self.state_size, activation="relu"))
        model.add(Dropout(0.2))  # Adding Dropout for regularization
        model.add(Dense(24, activation="relu"))
        model.add(Dropout(0.2))  # Adding Dropout for regularization
        model.add(Dense(self.action_size, activation="linear"))
        # Using legacy version of RMSprop as optimizer
        model.compile(loss=Huber(), optimizer=tf.keras.optimizers.legacy.RMSprop(learning_rate=self.learning_rate))
        return model

    def remember(self, state, action, reward, next_state, done):
        # Store experience in memory
        if len(self.memory) >= self.buffer_size:
            self.memory.pop(0)  # Remove the oldest experience
        self.memory.append((state, action, reward, next_state, done))

    def act(self, state):
        # Select action based on epsilon-greedy policy
        if np.random.rand() <= self.epsilon:
            return rnd.randint(0, self.action_size - 1)
        act_values = self.model.predict(np.expand_dims(state, axis=0))
        return np.argmax(act_values[0])

    def replay(self, batch_size):
        # Train the model with random experiences from memory
        self._train_model(batch_size)
        # Decay epsilon
        self._update_epsilon()
        # Update target model at set intervals
        self._update_target_model()
        # Adjust parameters a, b, d, and e based on action
        self._adjust_parameters()

    def _train_model(self, batch_size):
        # Train model with random experiences using Double Q-Learning
        minibatch = rnd.sample(self.memory, min(batch_size, len(self.memory)))
        for state, action, reward, next_state, done in minibatch:
            target = reward
            if not done:
                # Double Q-Learning: use model to select action and target model to evaluate it
                action_from_model = np.argmax(self.model.predict(np.expand_dims(next_state, axis=0))[0])
                target = reward + self.gamma * self.target_model.predict(np.expand_dims(next_state, axis=0))[0][action_from_model]
            target_f = self.model.predict(np.expand_dims(state, axis=0))
            target_f[0][action] = target
            # Train the model with the updated batch
            self.model.train_on_batch(np.expand_dims(state, axis=0), target_f)

        # Smart reduction of epsilon
        if self.epsilon > self.epsilon_min:
            self.epsilon *= self.epsilon_decay

    def _update_epsilon(self):
        # Update epsilon for exploration decay
        if self.epsilon > self.epsilon_min:
            self.epsilon *= self.epsilon_decay

    def _update_target_model(self):
        # Update the target model weights
        self.train_step += 1
        if self.train_step % self.target_update_freq == 0:
            self.update_target_model()

    def _adjust_parameters(self):
        # Adjusting parameters based on the last state and chosen action
        state = self.memory[-1][0]  # Assuming adjustment based on the last state
        action = self.act(state)
        #print(action)
        if 0 <= action < 10:
            self.a += (action - 5) * 0.01
        elif 10 <= action < 20:
            self.b += (action - 15) * 0.01
        elif 20 <= action < 30:
            self.d += (action - 25) * 0.01
        else:
            self.e += (action - 35) * 0.01
            
    def update_target_model(self):
        # Synchronize target model weights with the main model
        self.target_model.set_weights(self.model.get_weights())
    
# Machine Learning class
class ML:
    def __init__(self):
        # Read CSV file for data
        data = pd.read_csv("../../DATASET/mitbih_train.csv", header=None)
        test_data = pd.read_csv(
            "../../DATASET/mitbih_test.csv", header=None
        )  # Test dataset
        c0 = data[data[187] == 0].sample(1000)
        c1 = data[data[187] == 1].sample(1000)
        c2 = data[data[187] == 2].sample(1000)
        c3 = data[data[187] == 3]
        c4 = data[data[187] == 4].sample(1000)
        self.data = pd.concat([c0, c1, c2, c3, c4])
        self.test_data = test_data

    def load_data(self, selected_features):
        # Use features given by MH
        all_features = list(self.data.columns)
        selected_column_names = [
            all_features[i]
            for i in range(len(selected_features))
            if selected_features[i] == 1
        ]
        X_train = self.data[selected_column_names].values
        y_train = self.data.iloc[:, -1].values.astype(int)
        return X_train, y_train

    def load_test_data(self, selected_features):
        # Load test data based on selected features
        all_features = list(self.test_data.columns)
        selected_column_names = [
            all_features[i]
            for i in range(len(selected_features))
            if selected_features[i] == 1
        ]
        X_test = self.test_data[selected_column_names].values
        y_test = self.test_data.iloc[:, -1].values.astype(int)
        return X_test, y_test


# SVM based ML class
class MLSVM(ML):
    def train(self, selected_features):
        # Train the SVM model with selected features
        X_train, y_train = self.load_data(selected_features)
        X_test, y_test = self.load_test_data(selected_features)

        # Data preprocessing
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)
        X_test = scaler.transform(X_test)

        # Training the SVM model
        model = SVC()
        model.fit(X_train, y_train)

        # Evaluating the model
        y_pred = model.predict(X_test)

        acc = accuracy_score(y_test, y_pred)
        f1 = f1_score(y_test, y_pred, average="weighted")

        return [f1, acc]

# Random Forest based ML class
class MLRandomForest(ML):
    def train(self, selected_features):
        # Train the Random Forest model with selected features
        X_train, y_train = self.load_data(selected_features)
        X_test, y_test = self.load_test_data(selected_features)

        # Data preprocessing
        scaler = StandardScaler()
        X_train = scaler.fit_transform(X_train)
        X_test = scaler.transform(X_test)

        # Training the Random Forest model
        model = RandomForestClassifier()
        model.fit(X_train, y_train)

        # Evaluating the model
        y_pred = model.predict(X_test)
        
        acc = accuracy_score(y_test, y_pred)
        f1 = f1_score(y_test, y_pred, average='weighted')
        
        return [f1, acc]

# Orca class for optimization
class Orca:
    def __init__(self):
        # Initialize orca with random position and velocity
        self.dimension = 187
        self.position = []
        self.velocity = []
        self.metrics = []
        for j in range(self.dimension):
            self.position.append(rnd.randint(0, 1))
            self.velocity.append(rnd.uniform(0, 1))

    def isBetterThan(self, o):
        # Compare fitness of current orca with another
        return self.fitness() >= o.fitness()

    def countCols(self):
        # Count the number of selected columns
        return sum(self.position)

    def fitness(self):
        # Calculate fitness based on f1 score, accuracy, and number of columns used
        z1 = self.metrics[0]  # Maximize f1 score -> [0: bad, 1: good]
        z2 = self.metrics[1]  # Maximize acc -> [0: bad, 1: good]
        z3 = (
            self.countCols() / self.dimension
        )  # Minimize number of columns used -> [0: good, 1: bad]
        return z1 * 0.4 + z2 * 0.4 - z3 * 0.2 + 0.2

    def train(self, ml):
        # Train ML model and update metrics
        self.metrics = ml.train(self.position)

    # Chase movement of the orca
    def moveChaseOne(self, g, q, F, M, a, b, d, e):
        # Print values of a, b, d, and e
        print(f"a: {a}, b: {b}, d: {d}, e: {e}")
        for j in range(self.dimension):
            c = 1 - b
            if rnd.uniform(0, 1) > q:
                self.velocity[j] = a * (d * g.position[j] - F * (b * M[j] + c * self.position[j]))
            else:
                self.velocity[j] = e * (g.position[j] - self.position[j])

            # Update position based on velocity and logistic function
            self.position[j] = 1 if 1 / (1 + math.pow(math.e, -(self.position[j] + self.velocity[j]))) > rnd.uniform(0, 1) else 0

    # Second type of chase movement
    def moveChaseTwo(self, j1, j2, j3, u):
        for j in range(self.dimension):
            # Update position based on other orcas' positions
            self.position[j] = 1 if 1 / (1 + math.pow(math.e, -(j1.position[j] + u * (j2.position[j] - j3.position[j])))) else 0

    # Attack movement of the orca
    def moveAttack(self, co, o1, o2, o3, o4, j1, j2, j3):
        for j in range(self.dimension):
            # Calculate new velocity based on positions of other orcas
            v1 = (o1.position[j] + o2.position[j] + o3.position[j] + o4.position[j]) / 4 - co.position[j]
            v2 = (j1.position[j] + j2.position[j] + j3.position[j]) / 3 - self.position[j]
            self.velocity[j] = co.position[j] + rnd.uniform(0, 2) * v1 + rnd.uniform(-2.5, 2.5) * v2

    # String representation of orca fitness
    def __str__(self) -> str:
        return f"fit: {self.fitness()} <= => f1:{self.metrics[0]} acc:{self.metrics[1]} col:{self.countCols()/self.dimension}"

    # Copy attributes from another orca
    def copy(self, o):
        for j in range(o.dimension):
            self.position = o.position.copy()
            self.velocity = o.velocity.copy()
            self.metrics = o.metrics.copy()

# OPA Algorithm class
class OPA:
    def __init__(self):
        # Initialize parameters 
        self.nOrcas = 10
        self.T = 100
        self.p = 0.6
        self.q = 0.9
        self.F = 2

        # Initialize orcas and the best orca
        self.orcas = []
        self.gBest = Orca()
        # Initialize DQL with state and action sizes
        self.dql = DQL(
            5, 40
        )  

    def solve(self, ml, iteration):
        # Solve the optimization problem with given ML model and iteration count
        self.rnd(ml)
        self.run(ml, iteration)

    def rnd(self, ml):
        # Randomly initialize orcas and train them
        for i in range(self.nOrcas):
            self.orcas.append(Orca())
            self.orcas[i].train(ml)

        # Copy the best orca
        self.gBest.copy(self.orcas[0])
        for i in range(1, self.nOrcas):
            if self.orcas[i].isBetterThan(self.gBest):
                self.gBest.copy(self.orcas[i])
        print("t:0")
        self.bestToConsole()

    def run(self, ml, iteration):
        # Run the optimization process for a given number of iterations
        t = 1
        self.all_results = []  # Clear results at the start of each run
        # Initialize previous fitness and diversity
        fitness_anterior = 0
        diversidad_anterior = Utils.calculate_diversity(self.orcas)
        # Initialize previous f1 and accuracy
        f1_anterior = self.gBest.metrics[0]
        accuracy_anterior = self.gBest.metrics[1]
        self.all_results = []  # Clear results at the start of each run
        
        while t <= self.T:
            # Current state for reinforcement learning including performance metrics and diversity
            state = np.array([
                self.gBest.fitness(),
                Utils.calculate_diversity(self.orcas),
                self.gBest.countCols() / self.gBest.dimension,  # Percentage of selected columns
                self.gBest.metrics[0],  # F1 score
                self.gBest.metrics[1]   # Accuracy
            ])
            self.f1_anterior = self.gBest.metrics[0]
            self.accuracy_anterior = self.gBest.metrics[1]
            
            o = Orca()
            j1, j2, j3 = self.choose()
            if rnd.uniform(0, 1) > self.p:
                # Orca chase movement
                for i in range(self.nOrcas):
                    M = self.average()
                    self.orcas[i].moveChaseOne(self.gBest, self.q, self.F, M, self.dql.a, self.dql.b, self.dql.d, self.dql.e)

                # Dynamic adjustment based on time
                u = 2 * (rnd.uniform(0, 1) - 0.5) * (self.T - t) / self.T
                for i in range(self.nOrcas):
                    o.copy(self.orcas[i])
                    o.moveChaseTwo(j1, j2, j3, u)
                    o.train(ml)

                    # Update orcas and the best orca
                    if o.isBetterThan(self.orcas[i]):
                        self.orcas[i].copy(o)

                    if self.orcas[i].isBetterThan(self.gBest):
                        self.gBest.copy(self.orcas[i])
            else:
                # Orca attack movement
                for i in range(self.nOrcas):
                    o.copy(self.orcas[i])
                    o.moveAttack(self.orcas[i], self.find_orca(1), self.find_orca(2), self.find_orca(3), self.find_orca(4), j1, j2, j3)
                    
            # Calculate current diversity and change in diversity
            diversidad_actual = Utils.calculate_diversity(self.orcas)
            diversity_change = diversidad_actual - diversidad_anterior
            
            # Calculate reward based on performance and diversity changes
            reward = self.calculate_reward(diversity_change, f1_anterior, accuracy_anterior)

            # Update state and train DQL model
            action = self.dql.act(state)
            next_state = self.update_state()
            done = False  # Indicator of termination condition
            self.dql.remember(state, action, reward, next_state, done)
            self.dql.replay(32)  # Train DQL model with batch size

            # Prepare for next iteration
            fitness_anterior = self.gBest.fitness()
            diversidad_anterior = diversidad_actual
            f1_anterior = self.gBest.metrics[0]
            accuracy_anterior = self.gBest.metrics[1]
            state = next_state
            
            print(f"t:{t}")
            self.bestToConsole()
            diversidad_actual = Utils.calculate_diversity(self.orcas)
            print(f"Diversity after iteration {t}: {diversidad_actual}")

            if t == 50:
                # Save values to the list all_results instead of a file
                self.all_results.append(
                    [
                        self.gBest.fitness(),
                        self.gBest.metrics[0],
                        self.gBest.metrics[1],
                        self.gBest.countCols() / self.gBest.dimension,
                        1-(self.gBest.countCols() / self.gBest.dimension),
                        diversidad_actual,
                    ]
                )
            t += 1

    def find_orca(self, n):
        # Find an orca based on fitness ranking
        orcas_aux = self.orcas.copy()
        n_orcas = len(orcas_aux)

        # Sort orcas based on fitness
        for i in range(n_orcas - 1):
            for j in range(n_orcas - i - 1):
                if not orcas_aux[j].isBetterThan(orcas_aux[j + 1]):
                    orcas_aux[j], orcas_aux[j + 1] = orcas_aux[j + 1], orcas_aux[j]

        return orcas_aux[n - 1]
    
    # Helper function to calculate reward
    def calculate_reward(self, diversity_change, f1_anterior, accuracy_anterior):
        # Calculate changes in f1 and accuracy
        f1_change = self.gBest.metrics[0] - f1_anterior
        accuracy_change = self.gBest.metrics[1] - accuracy_anterior

        # Weights for reward components
        diversity_weight = 0.5
        column_weight = 0.3
        f1_weight = 0.1
        accuracy_weight = 0.1

        # Total reward calculation
        reward = (diversity_weight * diversity_change +
                  f1_weight * f1_change +
                  accuracy_weight * accuracy_change)
        
        # Bonus for having less than 30% of columns selected
        column_bonus = column_weight if (self.gBest.countCols() / self.gBest.dimension) < 0.3 else 0
        reward += column_bonus
    
        return reward

    # Helper function to update state
    def update_state(self):
        return np.array([
            self.gBest.fitness(),
            Utils.calculate_diversity(self.orcas),
            self.gBest.countCols() / self.gBest.dimension,
            self.gBest.metrics[0],
            self.gBest.metrics[1]
        ])     
            

    def save_results(self, ml):
        try:
            # Change filename according to the ML model
            filename = f'{ml.__class__.__name__}SOLO_OPA.xlsx'
            print(f"Trying to save in: {filename}")  # Debug

            # Create a DataFrame with values for each column
            df = pd.DataFrame(self.all_results, columns=['Fitness', 'F1 Score', 'Accuracy', 'Selected Columns', 'Reduccion', 'Diversidad'])

            # Check if the file exists
            if os.path.exists(filename):
                # Load existing file
                wb = load_workbook(filename)
                ws = wb.active
            else:
                # Create a new file and add headers
                wb = Workbook()
                ws = wb.active
                ws.append(['Fitness', 'F1 Score', 'Accuracy', 'Selected Columns','Reduccion','Diversidad'])

            # Add values to the end of the file
            for row in df.values.tolist():
                ws.append(row)

            # Save the file
            wb.save(filename)
            print(f"Data saved in: {filename}")  # Debug

        except Exception as e:
            print(f"Error saving data: {e}")    
            
    def average(self):
        # Calculate average position of orcas
        M = [0] * self.orcas[0].dimension
        for j in range(self.orcas[0].dimension):
            for i in range(self.nOrcas):
                M[j] = M[j] + self.orcas[i].position[j]
        for j in range(self.orcas[0].dimension):
            M[j] = M[j] / self.nOrcas
        return M

    def choose(self):
        # Choose three distinct orcas randomly
        while True:
            oj = rnd.randint(0, self.nOrcas - 1)
            ok = rnd.randint(0, self.nOrcas - 1)
            ol = rnd.randint(0, self.nOrcas - 1)
            if oj != ok and oj != ol and ol != ok:
                break
        return self.orcas[oj], self.orcas[ok], self.orcas[ol]

    def bestToConsole(self):
        # Print the best orca to the console
        print(" -- Best --")
        print(f"{self.gBest}")


